/*
  ALCATEL SpeedTouch USB modem library utility
  Copyright (C) 2001 Benoit PAPILLAULT
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Author   : Benoit PAPILLAULT <benoit.papillault@free.fr>
  Creation : ???

*/

#ifndef pool_h
#define pool_h

struct buffer
{
  /* a pointer to the memory allocated for this buffer */
  void * buf;

  /* size of the buffer allocated : the same size for all the buffer */
  int    size;

  /* =0 is the entry is free, =1 if it is in use by someone */
  int    allocated;
};

struct pool
{
  /* pointer to an array of 'buffer' */
  struct buffer * buf_list;

  /* first free buffer */
  int first_free;

  /* number of elements in the 'buf_list' array */
  int             buf_nb;

  /* only a big chunk of memory is allocated */
  void		* big_chunk;
};

/* All the following functions returns 1 in case of sucess
   and 0 in case of failure */

/* buffer_pool_init : 'nb' is the number of buffer to allocate
   and 'size' if the size in bytes of each buffer */
extern int buffer_pool_init(struct pool *p, int nb,int size);

/* buffer_pool_alloc : is used to reserve a buffer in the pool.
   a pointer to the 'buffer' allocated is returned in 'tbuf'
*/
extern int buffer_pool_alloc(struct pool *p, struct buffer ** tbuf);

/* buffer_pool_free : the 'buffer' is made available to the pool */
extern int buffer_pool_free(struct pool *p, struct buffer * tbuf);

/* search for a matching buffer containing 'ptr'. returns NULL if not found */
extern struct buffer * buffer_pool_search(struct pool *p, void * ptr);

#endif
