/*
  ALCATEL SpeedTouch USB modem utility : PPPoA implementation (2nd edition)
  Copyright (C) 2001 Benoit PAPILLAULT
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Author : Benoit PAPILLAULT <benoit.papillault@free.fr>
  Creation : 21/03/2001

  This program is designed to work under pppd, with the option "pty".

*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <time.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <termios.h>		/* N_HDLC & TIOCSETD */
#include <sys/resource.h>	/* setpriority() */
#include <string.h>
#if defined (__linux__)
#include <sched.h>		/* for sched_setscheduler */
#endif
#include <limits.h>		/* for LONG_MAX */
#include <sys/types.h>
#include <sys/socket.h>
#include <stdarg.h>

/* my USB library */
#include "pusb.h"

/* small semaphore lib. */
#include "smallsem.h"		/* provides with minimal semaphore/mutex fonctions */
#define	LOG_SEMKEY	0xCAFEDECA	/* "unique" key of the log's mutex */
int log_mutex=0;		/* exclusive acces when outputing to stdout */

/* for ident(1) command */
static char id[] =
	"@(#) $Id: pppoa2.c,v 1.9 2001/08/14 00:19:39 rogler Exp $";

/* USB level */
#define DATA_TIMEOUT 1000

#include "modem.h"

/* ATM level : adapt according your ADSL provider settings */
int my_vpi = -1;
int my_vci = -1;

#define CELL_SIZE 53
#define CELL_HDR  05
#define CELL_DATA 48

/* define for the size of buffer */
#define BUFFER_SIZE (64 * 1024)
#define HDLC_HEADER_SIZE 2

/*
* there is two buffers :
*  1) ppp read -> aal5 write
*  2) aal5_read -> ppp write
*     this buffer has 2 octet more, which are used for HDLC header
*/

/* so there is no more buffer copies between ppp and aal5 */
unsigned char ppp_read_buf[BUFFER_SIZE + HDLC_HEADER_SIZE];
unsigned char aal5_read_buf[BUFFER_SIZE + HDLC_HEADER_SIZE];
unsigned char *ppp_write_buf = &aal5_read_buf[HDLC_HEADER_SIZE];
unsigned char *aal5_write_buf = &ppp_read_buf[HDLC_HEADER_SIZE];

/*
* work buffer for aal5_write function
* used when generating the atm cells
*/
unsigned char aal5_work_buf[1367 * 53];

/* PPP level : DO NOT MODIFY */
#define syncHDLC 1

/*global parameters*/
int   verbose = 0;
int   usb_sync  = 1;

pid_t this_process;		/*always the current   process pid*/
pid_t parent_process;		/*always the parent    process pid*/
pid_t handler87_process;	/*always the handler87 process pid*/

#define CRC32_REMAINDER CBF43926
#define CRC32_INITIAL 0xffffffff

#define CRC32(c,crc) (crc32tab[((size_t)(crc>>24) ^ (c)) & 0xff] ^ (((crc) << 8)))

unsigned long crc32tab[256] = {

	0x00000000L, 0x04C11DB7L, 0x09823B6EL, 0x0D4326D9L,
	0x130476DCL, 0x17C56B6BL, 0x1A864DB2L, 0x1E475005L,
	0x2608EDB8L, 0x22C9F00FL, 0x2F8AD6D6L, 0x2B4BCB61L,
	0x350C9B64L, 0x31CD86D3L, 0x3C8EA00AL, 0x384FBDBDL,
	0x4C11DB70L, 0x48D0C6C7L, 0x4593E01EL, 0x4152FDA9L,
	0x5F15ADACL, 0x5BD4B01BL, 0x569796C2L, 0x52568B75L,
	0x6A1936C8L, 0x6ED82B7FL, 0x639B0DA6L, 0x675A1011L,
	0x791D4014L, 0x7DDC5DA3L, 0x709F7B7AL, 0x745E66CDL,
	0x9823B6E0L, 0x9CE2AB57L, 0x91A18D8EL, 0x95609039L,
	0x8B27C03CL, 0x8FE6DD8BL, 0x82A5FB52L, 0x8664E6E5L,
	0xBE2B5B58L, 0xBAEA46EFL, 0xB7A96036L, 0xB3687D81L,
	0xAD2F2D84L, 0xA9EE3033L, 0xA4AD16EAL, 0xA06C0B5DL,
	0xD4326D90L, 0xD0F37027L, 0xDDB056FEL, 0xD9714B49L,
	0xC7361B4CL, 0xC3F706FBL, 0xCEB42022L, 0xCA753D95L,
	0xF23A8028L, 0xF6FB9D9FL, 0xFBB8BB46L, 0xFF79A6F1L,
	0xE13EF6F4L, 0xE5FFEB43L, 0xE8BCCD9AL, 0xEC7DD02DL,
	0x34867077L, 0x30476DC0L, 0x3D044B19L, 0x39C556AEL,
	0x278206ABL, 0x23431B1CL, 0x2E003DC5L, 0x2AC12072L,
	0x128E9DCFL, 0x164F8078L, 0x1B0CA6A1L, 0x1FCDBB16L,
	0x018AEB13L, 0x054BF6A4L, 0x0808D07DL, 0x0CC9CDCAL,
	0x7897AB07L, 0x7C56B6B0L, 0x71159069L, 0x75D48DDEL,
	0x6B93DDDBL, 0x6F52C06CL, 0x6211E6B5L, 0x66D0FB02L,
	0x5E9F46BFL, 0x5A5E5B08L, 0x571D7DD1L, 0x53DC6066L,
	0x4D9B3063L, 0x495A2DD4L, 0x44190B0DL, 0x40D816BAL,
	0xACA5C697L, 0xA864DB20L, 0xA527FDF9L, 0xA1E6E04EL,
	0xBFA1B04BL, 0xBB60ADFCL, 0xB6238B25L, 0xB2E29692L,
	0x8AAD2B2FL, 0x8E6C3698L, 0x832F1041L, 0x87EE0DF6L,
	0x99A95DF3L, 0x9D684044L, 0x902B669DL, 0x94EA7B2AL,
	0xE0B41DE7L, 0xE4750050L, 0xE9362689L, 0xEDF73B3EL,
	0xF3B06B3BL, 0xF771768CL, 0xFA325055L, 0xFEF34DE2L,
	0xC6BCF05FL, 0xC27DEDE8L, 0xCF3ECB31L, 0xCBFFD686L,
	0xD5B88683L, 0xD1799B34L, 0xDC3ABDEDL, 0xD8FBA05AL,
	0x690CE0EEL, 0x6DCDFD59L, 0x608EDB80L, 0x644FC637L,
	0x7A089632L, 0x7EC98B85L, 0x738AAD5CL, 0x774BB0EBL,
	0x4F040D56L, 0x4BC510E1L, 0x46863638L, 0x42472B8FL,
	0x5C007B8AL, 0x58C1663DL, 0x558240E4L, 0x51435D53L,
	0x251D3B9EL, 0x21DC2629L, 0x2C9F00F0L, 0x285E1D47L,
	0x36194D42L, 0x32D850F5L, 0x3F9B762CL, 0x3B5A6B9BL,
	0x0315D626L, 0x07D4CB91L, 0x0A97ED48L, 0x0E56F0FFL,
	0x1011A0FAL, 0x14D0BD4DL, 0x19939B94L, 0x1D528623L,
	0xF12F560EL, 0xF5EE4BB9L, 0xF8AD6D60L, 0xFC6C70D7L,
	0xE22B20D2L, 0xE6EA3D65L, 0xEBA91BBCL, 0xEF68060BL,
	0xD727BBB6L, 0xD3E6A601L, 0xDEA580D8L, 0xDA649D6FL,
	0xC423CD6AL, 0xC0E2D0DDL, 0xCDA1F604L, 0xC960EBB3L,
	0xBD3E8D7EL, 0xB9FF90C9L, 0xB4BCB610L, 0xB07DABA7L,
	0xAE3AFBA2L, 0xAAFBE615L, 0xA7B8C0CCL, 0xA379DD7BL,
	0x9B3660C6L, 0x9FF77D71L, 0x92B45BA8L, 0x9675461FL,
	0x8832161AL, 0x8CF30BADL, 0x81B02D74L, 0x857130C3L,
	0x5D8A9099L, 0x594B8D2EL, 0x5408ABF7L, 0x50C9B640L,
	0x4E8EE645L, 0x4A4FFBF2L, 0x470CDD2BL, 0x43CDC09CL,
	0x7B827D21L, 0x7F436096L, 0x7200464FL, 0x76C15BF8L,
	0x68860BFDL, 0x6C47164AL, 0x61043093L, 0x65C52D24L,
	0x119B4BE9L, 0x155A565EL, 0x18197087L, 0x1CD86D30L,
	0x029F3D35L, 0x065E2082L, 0x0B1D065BL, 0x0FDC1BECL,
	0x3793A651L, 0x3352BBE6L, 0x3E119D3FL, 0x3AD08088L,
	0x2497D08DL, 0x2056CD3AL, 0x2D15EBE3L, 0x29D4F654L,
	0xC5A92679L, 0xC1683BCEL, 0xCC2B1D17L, 0xC8EA00A0L,
	0xD6AD50A5L, 0xD26C4D12L, 0xDF2F6BCBL, 0xDBEE767CL,
	0xE3A1CBC1L, 0xE760D676L, 0xEA23F0AFL, 0xEEE2ED18L,
	0xF0A5BD1DL, 0xF464A0AAL, 0xF9278673L, 0xFDE69BC4L,
	0x89B8FD09L, 0x8D79E0BEL, 0x803AC667L, 0x84FBDBD0L,
	0x9ABC8BD5L, 0x9E7D9662L, 0x933EB0BBL, 0x97FFAD0CL,
	0xAFB010B1L, 0xAB710D06L, 0xA6322BDFL, 0xA2F33668L,
	0xBCB4666DL, 0xB8757BDAL, 0xB5365D03L, 0xB1F740B4L

};

unsigned int
calc_crc(unsigned char *mem, int len, unsigned int initial)
{

	unsigned int crc;

	crc = initial;

	for(; len; mem++, len--)
		crc = CRC32(*mem, crc);


	return (crc);

}

/*
* Logs a message if verbose level is higher than
* the minlevel required
*/
void report( int minlevel, int perror_needed, char *format, ...)
{

	va_list ap;
	time_t tps;

	if( verbose > (minlevel-1))
	{

		va_start(ap, format);

		time(&tps);

		sem_P(log_mutex);	/* Enter critical section */

		printf("< pid=%d > [%.24s] ", this_process, ctime(&tps));
		vprintf(format, ap);

		if(perror_needed)
			perror("Reason");

		sem_V(log_mutex);	/* leave critical section */

		va_end(ap);

	}

	return;

}

/*
* display a convenient string about the current time.
* the line printed is not \n terminated
*/

void
print_time()
{

	struct timeval tv;

	gettimeofday(&tv, NULL);

	sem_P(log_mutex);	/* enter critical section */

	printf("[%.24s %6ld]", ctime(&tv.tv_sec), tv.tv_usec);

	sem_V(log_mutex);	/* leave critical section */

}

/*
 * Note: It is not necessary to protect this printf
 * with a mutex, done from dump() bellow.
 */
void
print_char(unsigned char c)
{

	if(c >= ' ' && c < 0x7f)
		printf("%c", c);
	/*
	else if (c=='\r' || c=='\n' || c=='\t' || c=='\b')
	printf("%c",c);
	*/
	else
		printf(".");

}

void
dump(unsigned char *buf, int len)
{

	int i, j;

	sem_P(log_mutex);	/* enter critical section */

	for(i = 0; i < len; i += 16)
	{

		for(j = i; j < len && j < i + 16; j++)
			printf("%02x ", buf[j]);

		for(; j < i + 16; j++)
			printf("   ");

		for(j = i; j < len && j < i + 16; j++)
			print_char(buf[j]);

		printf("\n");

	}
	
	printf("\n");

	sem_V(log_mutex);   /* leave cricital section */

}

/*
* ppp_read returns a pointer to a static buffer.
* it returns the number of bytes read or -1 in case of errors.
* 0 should indicates that pppd exited (not tested).
*/

int
ppp_read(int fd)
{

	/*
	* we must handle HDLC framing, since this is what pppd
	* sends to us. I use some code from rp-pppoe
	*/

	int r;

	for(;;)
	{

		do {
			r = read(fd, ppp_read_buf, BUFFER_SIZE);
		} while(r < 0 && errno == EINTR);

		if(r < 0)
			report(0, 1, "Error reading from pppd\n");

		/*Debug message*/
		report(2, 0, "Read from ppp, len = %d\n", r);
		if(verbose>1) dump(ppp_read_buf, r);

		if(r <= 0)
			return r;

		/* supress leading 2 bytes */

		if(syncHDLC)
		{

			if(r < 3)
			{
				report(0, 0, "Read from ppp short on data, r=%d => ignored\n", r);
				continue;
			}

			if(r == BUFFER_SIZE)
			{
				report(0, 0, "Read from ppp too much data, r=%d => ignored\n", r);
				continue;
			}

			return(r - HDLC_HEADER_SIZE);

		}
		else
		{

			report(0, 0, "async HDLC not handled\n");

		}

		break;

	}

}

#define NOBUF_RETRIES 5

int
ppp_write(int fd, int n)
{
	int r;
	static int errs = 0;
	int retries = 0;

	if(syncHDLC)
	{

		report(2, 0, "writing to ppp, len = %d\n", n);
		if(verbose>1) dump(aal5_read_buf, n);

	retry:

		if((r = write(fd, aal5_read_buf, n)) < 0)
		{

			/*
			* We sometimes run out of mbufs doing simultaneous
			* up- and down-loads on FreeBSD.  We should find out
			* why this is, but for now...
			*/

			if(errno == ENOBUFS)
			{

				errs++;

				if((errs < 10 || errs % 100 == 0))
					report(0, 0, "ppp_write: %d ENOBUFS errors\n", errs);

				if(retries++ < NOBUF_RETRIES)
				{
					/* retry after delay */
					usleep(1000);
					goto retry;
				}
				else
				{
					/* throw away the packet */
					return(0);
				}

			}
			else
			{
				report(0, 1, "Error writing to pppd\n");
			}
		}
		return(r);
	}
	else
	{
		report(0, 0, "async HDLC not handled\n");
	}

	return(-1);
}

/* 
* parse an ATM cell and try to complete (or make a new one)
* AAL5 frame. If succes, the frame is sent to pppd. If error, -1 is returned
* (0 otherwise)
*/

int
aal5_read(unsigned char *cell_buf, size_t count,
					unsigned int vpi, unsigned int vci, unsigned int pti, int fdout)
{
	static int len = 0;

	int ret;
	unsigned int real_len;

	if(vpi != my_vpi || vci != my_vci)
	{

		report(1, 0, "vpi.vci mismatch (%d/%d instead of %d/%d) ... ignored!\n",
				vpi, vci, my_vpi, my_vci);
		if(verbose>1) dump(cell_buf, CELL_DATA);

		return(0);
	
	}

	/*
	* We should add a check for OAM cells, which should be done before
	* copying the cell's content to our buffer
	*/

	if(len + CELL_DATA < BUFFER_SIZE)
	{

		memcpy(ppp_write_buf + len, cell_buf, CELL_DATA);
		len += CELL_DATA;

	}
	else
	{
		report(1, 0, "aal5_read: buffer overflow (len=%d) => reseting buffer\n", len);

		len = 0;
		return(0);

	}

	/* check if PTI=1, which indicated last cell of an AAL5 frame */
	if((pti & 2) == 0)
		return(0);

	real_len = (ppp_write_buf[len - 6] << 8) | ppp_write_buf[len - 5];

	report(2, 0, "aal5_read: len=%d => real_len=%d\n", len, real_len);

	aal5_read_buf[0] = 0xff;	//FRAME_ADDR;
	aal5_read_buf[1] = 0x03;	//FRAME_CTRL;

	if((ret = ppp_write(fdout, real_len + HDLC_HEADER_SIZE)) < 0)
		return(ret);

	len = 0;

	return(0);

}

/*
* cell_read : is used to decode an individual cell and to try to make
* an AAL5 frame.
*/

int
cell_read(unsigned char *cellbuf, size_t count, int fdout)
{

	unsigned int vpi, vci, pti;

	/* decoding & printing of the cell */
	vpi = ( cellbuf[0]         <<  4) | (cellbuf[1] >> 4);
	vci = ((cellbuf[1] & 0x0f) << 12) | (cellbuf[2] << 4) | (cellbuf[3] >> 4);
	pti = ( cellbuf[3] & 0x0f);

	/* filling our AAL5 frame */
	return(aal5_read(cellbuf + CELL_HDR, CELL_DATA, vpi, vci, pti, fdout));
}

int
cell_write(unsigned char *cellbuf,
					 unsigned char *buf, int n,
					 unsigned int vpi, unsigned int vci, unsigned int pti)
{

	unsigned char gfc = 0;
	unsigned char hec = 0xec;

	if(n != CELL_DATA)
		return(-1);

	/* normalize */
	vpi &= 0xff;
	vci &= 0xffff;
	pti &= 0x0f;

	cellbuf[0] = (gfc << 4) | (vpi >> 4);
	cellbuf[1] = (vpi << 4) | (vci >> 12);
	cellbuf[2] = (vci >> 4);
	cellbuf[3] = ((vci & 0xf) << 4) | pti;
	cellbuf[4] = hec;

	memcpy(cellbuf + CELL_HDR, buf, CELL_DATA);

	return(CELL_DATA);

}

int other_vpi = 0;
int other_vci = 0;

#define VCI_MAX 0xff
#define VPI_MAX 0xffff

void
next_vpi_vci()
{

	if(++other_vci >= VCI_MAX)
	{
		other_vci = 0;
		if(++other_vpi >= VPI_MAX)
			other_vpi = 0;
	}

}


/*
* Warning: (ok, tooks me half an hour to find my own buf)
* buf must be large enough to contains the whole AAL5 frame,
* with its layer.
*/

/* use ppp_read_buf and aal5_work_buf */
int
aal5_write(pusb_endpoint_t epdata, int n)
{

	/* add the trailer */
	int total_len = 48 * ((n + 8 + CELL_DATA - 1) / 48);
	unsigned int crc, pti = 0;

	/*
	* this buffer will contain all cells for an AAL5 frame,
	* so AAL5 frame is max 64K + 48 that's 1367 cells at max
	*/

	int ptr,ptr2;
	int ret;

	/* filling padding bytes with 0's */
	memset(aal5_write_buf + n, 0, total_len - n);

	aal5_write_buf[total_len - 6] = n >> 8;
	aal5_write_buf[total_len - 5] = n & 0xff;

	crc = ~calc_crc(aal5_write_buf, total_len - 4, -1);

	aal5_write_buf[total_len - 4] = crc >> 24;
	aal5_write_buf[total_len - 3] = crc >> 16;
	aal5_write_buf[total_len - 2] = crc >> 8;
	aal5_write_buf[total_len - 1] = crc;

	ptr = ptr2 = 0;

	while(total_len > 0)
	{

		if(total_len <= CELL_DATA)
			/* this is the last cell, so mark it */
			pti = 2;

		cell_write(aal5_work_buf + ptr, aal5_write_buf + ptr2, CELL_DATA, my_vpi, my_vci, pti);
		ptr += CELL_SIZE;
		ptr2 += CELL_DATA;

		/* send the same cell over another vpi/vci */

		/*
		printf("trying vpi=%u, vci=%u\n",other_vpi,other_vci);
		cell_write(aal5_work_buf+ptr,aal5_write_buf,CELL_DATA,other_vpi,other_vci,pti);
		ptr += CELL_SIZE;

		next_vpi_vci();
		*/

		total_len -= CELL_DATA;

	}

	/* Debug message */
	report(2, 0, "writing to usb, len = %d\n", ptr);
	if(verbose>1) dump(aal5_work_buf, ptr);

	ret = pusb_endpoint_write(epdata, aal5_work_buf, ptr, DATA_TIMEOUT);
	
	if(ret < 0)
		report(0, 1, "Error writing to usb\n");

	return(ret);
}

/*
* variable that will be used in several function
*   => so global
*/

int gfdout;

/* returns -1 in case of errors */

int
decode_usb_pkt(unsigned char *buf, int len)
{

	int r;

	/*Debug message*/
	report(2, 0, "read from usb, len = %d\n", len);
	if(verbose>1) dump(buf, len);

	/*
	* ... ??? ... 
	* I prefer to say nothing about the following lines
	*/

	if((len % CELL_SIZE) != 0)
	{

		int junk = len % CELL_SIZE;

		report(1, 0, "junk len = %d len%%%d = %d\n", len, CELL_SIZE, junk);

		len -= junk;
		buf += junk;
	}

	while(len >= CELL_SIZE)
	{

		if((r = cell_read(buf, CELL_SIZE, gfdout)) < 0)
			return(r);

		buf += CELL_SIZE;
		len -= CELL_SIZE;

	};

	if(len != 0)
		report(1, 0, "arf : len = %d\n", len);

	return(0);

}

/*
* endpoint 0x81, which is on interface 0 is used for handling 'interrupt'.
* We will handle it that way : keep reading on it!
*
* PS : not used
*
*/

void
handle_endpoint_81(pusb_endpoint_t ep_int)
{
	pid_t child_pid;

	if((child_pid = fork()) == 0)
	{

		unsigned char lbuf[64 * 53];
		int ret;

		for(;;)
		{
			ret = pusb_endpoint_read(ep_int, lbuf, sizeof(lbuf), 0);

			if(ret < 0)
			{
				report(0, 0, "Error reading interrupts\n");
				break;
			}

			report(2, 0, "received interrupts, len = %d\n", ret);
			if(verbose>1) dump(lbuf, ret);

		}

		_exit(0);
	}

}

/* 
* endpoint 0x87 is used for receiving ATM cells
*
* to handle synchronous read, we fork a new process
* which does an infinite loop, keeping reading on 
* endpoint EP_DATA_IN 
*/

void
handle_endpoint_87(pusb_endpoint_t epdata, int fdout)
{
	int ret;
	pid_t child_pid;

	/*
	other_process = getpid();
	*/

	if((child_pid = fork()) == 0)
	{

		unsigned char lbuf[64 * 53];

		/*
		* This is the children part of the fork()
		* child_pid doesn't mean nothing, just we are a child process
		* so we get our real pid
		*/

		this_process = handler87_process = getpid();

		sem_P(log_mutex);	/* Enter critical section */

		printf("* pid=%d * : I'm the children process, I handle the endpoint 0x87\n", this_process);

		sem_V(log_mutex);	/* Leave critical section */
		
		for(;;)
		{

			ret = pusb_endpoint_read(epdata, lbuf, sizeof(lbuf), 0);

			if(ret < 0)
			{
				report(0, 1, "Error reading from usb\n");
				break;
			}

			if(decode_usb_pkt(lbuf, ret) < 0)
			{
				report(0, 1, "Error decoding from usb");
				break;
			}
		}

		/*
		* this line is reached only if an error occurs in
		* communicating with usb or pppd
		*/

		/* first, we kill our father, which is useless */
		if(kill(parent_process, SIGINT) < 0)
			report(0, 1, "Error killing parent process\n");

		sem_destroy(log_mutex);	/* destroys the semaphore from the system */

		/* now we exit from this process too*/
		_exit(0);

	}
	else {

		/*
		* This is the parent part of the fork
		* child_pid is the pid of the forked process
		* we save it in handler87_process to be able to kill it
		* when a signal is handled by sighandler()
		*/

		handler87_process = child_pid;

	}

}

/* 
* endpoint 0x07 is used for sending ATM cells
*
* to handle synchronous write, we does an infinite
* loop, keeping writing on endpoint EP_DATA_OUT
*
* NB : this function must be called after the
*      handle_endpoint_87 which is a fork.
*      handle_endpoint_07 is an infinite loop
*      so if you enter it, you will never get
*      out ( just when there are errors)
*/

void
handle_endpoint_07(pusb_endpoint_t epdata, int fdin)
{
	/* this part still need a redesign ... */
	int r, n;

	for(;;)
	{

		n = ppp_read(fdin);

		if(n <= 0)
			break;

		r = aal5_write(epdata, n);

		if(r < 0)
			break;
	}

}

void
usage()
{

	fprintf(stderr, "pppoa2 version $Name: tar_gz_15082001 $\n");
	fprintf(stderr, "usage: pppoa2 [-v level] [-f logfile] -vpi val -vci val\n");
	fprintf(stderr, "  -v       : defines the verbosity level [0-2]\n");
	fprintf(stderr, "  -f       : defines the log filename to use ( Default /tmp/pppoa2.log )\n");
	fprintf(stderr, "  -vpi.vci : defined the vpi.vci that your provider is\n");
	fprintf(stderr, "             using, 8.35 for France, 0.38 for UK\n");
	fprintf(stderr, "  --help   : this message\n");
	exit(-1);

}

/*
* Die on a hangup, and propagate it to the handler_87 process
*/

void
sighandler(int signal)
{

	/*We kill the handler87 handler process*/
	if(kill(handler87_process, signal) < 0)
		report(0, 1, "Error killing handler87 process\n");

	report(0, 0, "Killed by a signal %d\n", signal);

	sem_destroy(log_mutex);	/* destroys the semaphore from the system */

	/* We exit safely*/
	_exit(1);
}

int
main(int argc, char *argv[])
{
	char   *logfile;
	int    fdin, fdout, log;
	int    i;
	time_t ourtime;

	pusb_device_t fdusb;
	pusb_endpoint_t ep_data;

	logfile = "/var/log/pppoa2.log";
	log = 0;

	for(i = 1; i < argc; i++)
	{

		if(strcmp(argv[i], "-v") == 0 && i + 1 < argc)
			verbose = atoi(argv[++i]);
		else if(strcmp(argv[i], "-vpi") == 0 && i + 1 < argc)
			my_vpi = atoi(argv[++i]);
		else if(strcmp(argv[i], "-vci") == 0 && i + 1 < argc)
			my_vci = atoi(argv[++i]);
		else if(strcmp(argv[i], "--help") == 0)
			usage();
		else if(strcmp(argv[i], "-f") == 0 && i + 1 < argc)
			logfile   = argv[++i];
		else 
			usage();
	}

	/*vpi & vci aren't set*/
	if(my_vpi == -1 || my_vci == -1)
		usage();

	/*Duplicate in and out fd*/
	fdin  = dup(STDIN_FILENO);
	fdout = dup(STDOUT_FILENO);

	/*Get process pid*/
	parent_process = this_process = getpid();

	/* Create the log */
	log = open(logfile, O_CREAT | O_RDWR | O_APPEND, 0600);

	/* Failed to create the log file*/
	if(log < 0)
	{
		perror(logfile);
		return(-1);
	}


	/*
	 * Check if our mutex remains from previous executions
	 * of pppoa2. If yes, destroy it. 
	 * Then create a fresh one and set its initial value to 1.
	 */
	if ((log_mutex=sem_get((key_t)LOG_SEMKEY))!=-1)
	{
		if (sem_destroy(log_mutex)==-1)
		{
			perror("sem_destroy");
			return(-1);
		}
	}
	if ((log_mutex=sem_create((key_t)LOG_SEMKEY))==-1)
	{
		perror("sem_create");
		return(-1);
	}
	if (sem_init(log_mutex, 1)==-1)
	{
		perror("sem_init");
		return(-1);
	}

	/*
	* we will redirect all standard streams to the log file
	* so we are sure to catch all messages from this program :)
	*/
	if(dup2(log, STDIN_FILENO) == -1)
	{
		write(log,"Error duplicating STDIN_FILENO\n", 31);
		return(-1);
	}
	if(dup2(log, STDOUT_FILENO) == -1)
	{
		write(log,"Error duplicating STDOUT_FILENO\n", 32);
		return(-1);
	}
	if(dup2(log, STDERR_FILENO) == -1)
	{
		write(log,"Error duplicating STDERR_FILENO\n", 32);
		return(-1);
	}

	/*We don't need log fd anymore*/
	if( close(log) == -1)
	{
		fprintf(stderr, "Error closing log fd\n");
		return(-1);
	}

	/*
	* No buffering on stdout & stderr
	* So the stdout and stderr are flushed
	* each time we write to them
	*/
	setbuf(stdout, NULL);
	setbuf(stderr, NULL);

	sem_P(log_mutex);	/* Enter critical section */

	/* Start the log file with the date */
	for(i=0; i<8 ; i++)
		printf("##########");

	printf("\n\nPPPoA2 version %s\n",VERSION);

	time(&ourtime);
	printf("Log started on %s\n", ctime(&ourtime));

	/* If level > 1, write the vpi,vci used */
	if(verbose>1) printf("Using vpi=%d, vci=%d\n", my_vpi, my_vci);

	/* To know what pid is the parent process */
	printf("* pid=%d * : I'm the parent   process, I handle the endpoint 0x07\n", parent_process);

	sem_V(log_mutex);	/* Leave critical section */

	/* Give information about the tty fds */
	report(1, 0, "pty descriptors : fdin=%d, fdout=%d\n", fdin, fdout);



	/* Increase priority of the pppoa process*/
	if(setpriority(PRIO_PROCESS, this_process, -20) < 0)
		report(1, 1,"setpriority failed\n");

#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
#define SOCKBUF (64*1024)
	{
		int sbuf, ss = sizeof(sbuf);

		if(getsockopt(fdin, SOL_SOCKET, SO_SNDBUF, &sbuf, &ss) == 0)
		{

			/*Debug message*/
			report(1, 0, "Increasing SNDBUF from %d to %d\n", sbuf, SOCKBUF);

			sbuf = SOCKBUF;

			if(setsockopt(fdin, SOL_SOCKET, SO_SNDBUF, &sbuf, ss) < 0)
				report(0, 1, "setsockopt failed\n");

		}

		if(getsockopt(fdin, SOL_SOCKET, SO_RCVBUF, &sbuf, &ss) == 0)
		{

			/*Debug message*/
			report(1, 0, "Increasing RCVBUF from %d to %d\n", sbuf, SOCKBUF);

			sbuf = SOCKBUF;

			if(setsockopt(fdin, SOL_SOCKET, SO_RCVBUF, &sbuf, ss) < 0)
				report(0, 1, "setsockopt failed\n");

		}

	}
#endif

	/*
	* Install HDLC line discipline on fdin if it is a tty and
	* the OS has such a thing.
	*/

	if(isatty(fdin))
	{
#ifdef N_HDLC
		int disc = N_HDLC;
		if(ioctl(fdin, TIOCSETD, &disc) < 0)
		{
			report(0, 0, "Error loading N_HDLC\n");
			return(-1);
		}
		report(2, 0, "N_HDLC line set up\n");

#elif defined TTYDISC
		int disc = TTYDISC;
		if(ioctl(fdin, TIOCSETD, &disc) < 0)
		{
			report(0, 0, "Error setting termios tty line discipline\n");
			return(-1);
		}
		report(2, 0, "TTYDISC line set up\n");

#endif
	}

	/*
	*  We search for the first USB device matching ST_VENDOR & ST_PRODUCT.
	*  usbdevfs must be mount on /proc/bus/usb (or you may change the path
	*  here, according to your config
	*/

	fdusb = pusb_search_open(ST_VENDOR, ST_PRODUCT);

	if(fdusb == NULL)
	{
		report(0, 0, "Where is this crappy modem ?!\n");
		return(-1);
	}

	/*Debug message*/
	report(2, 0, "Got the modem !\n");

	/* Initialize global variables */
	gfdout = fdout;

	/* We claim interface 1, where endpoints 0x07 & 0x87 are */
	if(pusb_claim_interface(fdusb, 1) < 0)
	{
		report(0, 0, "pusb_claim_interface 1 failed\n");
		return(-1);
	}

	ep_data = pusb_endpoint_open(fdusb, EP_DATA_OUT, O_RDWR);
	if(ep_data == NULL)
	{
		report(0, 0, "pusb_endpoint_open failed\n");
		return(-1);
	}

	/*
	* endpoint 0x87 is used for receiving ATM cells, this function will
	* fork a new process that will read ATM cells from the usb and send them
	* to pppd
        */

	handle_endpoint_87(ep_data, fdout);

	/* At least in FreeBSD, ppp may kill us with a hangup. */
	/* In Linux, pppd kill us with a term signal ! */
#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
	signal(SIGHUP , sighandler);
#elif defined(__linux__)
	signal(SIGTERM, sighandler);
#endif


	

	/*
	* endpoint 0x07 is used for sending ATM cells, this function
	* is an infinite loop and will return only on errors.
        */

	handle_endpoint_07(ep_data, fdin);

	/*
	*  we kill our child pid : this is to be sure that
	*  the USB device file will be available for other programs
	*/

	if(kill(handler87_process, SIGINT) < 0)
		report(0, 1, "Error killing child process %d\n", handler87_process);

	/* we release all the interface we'd claim before exiting */
	if(pusb_release_interface(fdusb, 1) < 0)
		report(0, 1,"pusb_release_interface failed\n");

	pusb_endpoint_close(ep_data);

	pusb_close(fdusb);

	sem_destroy(log_mutex);	/* destroys the semaphore from the system */

	return(0);

}
